from itertools import chain
from typing import Dict, List

from roboswag.generate.models.parameter import Parameter
from roboswag.generate.models.response import Response


class Endpoint:
    def __init__(
        self,
        method_name: str,
        http_method,
        url,
        summary: str,
        description: str,
        path_params: List[Parameter],
        headers: List[Parameter],
        query: List[Parameter],
        body: Parameter,
        responses: Dict[str, Response],
        exp_status: str = "200",
    ):
        self.method_name: str = method_name
        self.http_method = http_method
        self.url = self.prepare_path(url, path_params)
        self.summary: str = summary
        self.description: str = description
        self.path_params: List[Parameter] = path_params
        self.headers: List[Parameter] = headers
        self.query: List[Parameter] = query
        self.body: Parameter = body
        self.responses: Dict[str, Response] = responses
        self.exp_status = exp_status
        self.method_signature: str = self.get_python_method_signature()

    @staticmethod
    def prepare_path(url, path_params):
        for param in path_params:
            url = url.replace(f"{{{param.name}}}", f"{{{param.python_name}}}")
        return url

    def get_python_method_signature(self) -> str:
        max_line_length: int = 120
        args = [str(param) for param in chain(self.path_params, self.headers, self.query) if param]
        if self.body:
            args.append(f"{self.body}")
        args.append(f"exp_status={self.exp_status}")
        args.append(f"validate_schema=True")
        line = f"    def {self.method_name}(self"
        prefix = len(line) - 4
        last_index = len(args) - 1
        method_sig = ""
        for index, arg in enumerate(args):
            if not arg:
                continue
            if len(line) + len(arg) + 2 > max_line_length - (2 if last_index == index else 0):
                method_sig += line
                line = ",\n" + prefix * " " + arg
            else:
                line += f", {arg}"
        method_sig += line
        return method_sig



from roboswag.generate.models.utils import pythonify_name


class Parameter:
    def __init__(self, name: str, default=None, param_type=None, description=None, required=None, schema=None) -> None:
        self.name: str = name
        self.python_name: str = pythonify_name(name)
        self.default = default
        self.param_type = param_type
        self.description = description
        self.required = required
        self.schema = schema

    def __str__(self):
        if self.default is None:
            return self.python_name
        return f"{self.python_name}={self.default}"



import requests
import urllib3

from roboswag.auth import TokenHandler
from roboswag.logger import Logger
from roboswag.validate import Validate

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


class APIModel:
    def __init__(
        self,
        base_url,
        verify=False,
        headers=None,
        content_type="application/json",
        proxies=None,
        allow_redirects=True,
        authorization=TokenHandler,
    ):
        # Set headers from init too (or reuse auth)
        self.base_url = base_url
        self.session = requests.Session()
        self.session.verify = verify
        self.content_type = content_type
        if content_type is not None:
            self.session.headers = {"Content-Type": content_type}
        self.allow_redirects = allow_redirects
        if proxies is not None:  # TODO urllib have autodetect proxy - allow to use it
            self.session.proxies.update(proxies)
        self.authorization = authorization
        self.logger = Logger()
        self.validate = Validate(self.logger)
        if headers is not None:
            self.session.headers.update(headers)

    def send_request(self, method, url, status=None, headers=None, body=None, query=None, **kwargs):
        headers = self.trim_empty(headers)
        query = self.trim_empty(query)
        # auth = self.authorization(**kwargs) if self.authorization is not None else None
        content_type = kwargs.get("content-type", self.content_type)
        if content_type is not None:
            headers["Content-Type"] = content_type

        resp = self.session.request(
            method,
            url=self.base_url + url,
            headers=headers,
            json=body,
            params=query,
            # auth=auth,
            allow_redirects=self.allow_redirects,
        )
        # TODO quiet mode
        self.logger.log_request(resp)
        self.logger.log_response(resp)
        if status is not None:
            assert resp.status_code == status, f"Expected return status: {status} but received: {resp.status_code}"
        return resp

    def post(self, *args, **kwargs):
        # TODO handle files upload
        return self.send_request("POST", *args, **kwargs)

    def get(self, *args, **kwargs):
        return self.send_request("GET", *args, **kwargs)

    def put(self, *args, **kwargs):
        return self.send_request("PUT", *args, **kwargs)

    def delete(self, *args, **kwargs):
        return self.send_request("DELETE", *args, **kwargs)

    @staticmethod
    def trim_empty(dictionary):
        if dictionary is None:
            return {}
        return {key: value for key, value in dictionary.items() if value is not None}



class Response:
    def __init__(self, description: str, headers=None, schema=None):
        self.description = description
        self.headers = headers
        self.schema = schema

    def __repr__(self):
        return f"schema: {self.schema}"



"""
You can use your own auth class when defining your API endpoint handler:
    from webapi.auth import HTTPBasicAuth

    class SomeAPI(APICore):
        def __init__(self, base_url, *args, **kwargs):
            super().__init__(base_url, *args, authorization=HTTPBasicAuth, **kwargs)

See https://copdips.com/2019/04/creating-custom-python-request-auth-class.html for example how you can
define custom auth classes.
"""
from requests.auth import HTTPBasicAuth
from robot.libraries.BuiltIn import BuiltIn


class MissingParameter(ValueError):
    def __init__(self, name):
        super().__init__(f"Missing {name}. Set ${{{name}}} global variable or pass it as {name} named variable.")


def get_from_kwargs_or_robot(kwargs, name, missing_ok=False):
    value = kwargs.get(name)
    if value is None:
        value = BuiltIn().get_variable_value(f"${{{name}}}")
    if not missing_ok and value is None:
        raise MissingParameter(name) from None
    return value


class TokenHandler:
    def __init__(self, **kwargs):
        self.access_token = get_from_kwargs_or_robot(kwargs, "access_token")

    def __call__(self, *args, **kwargs):
        user = get_from_kwargs_or_robot(kwargs, "user")
        password = get_from_kwargs_or_robot(kwargs, "password")
        super().__init__(user, password)



import json
from pathlib import Path
from typing import Dict, Union

import jsonschema

from roboswag.validate.core import ValidateBase


class ValidateSchema(ValidateBase):
    def schema(self, to_validate, schema: Union[str, Path, Dict]):
        self.logger.info("Validating schema...")
        if isinstance(schema, (str, Path)):
            with open(schema) as fp:
                schema = json.load(fp)
        self.logger.debug(f"Schema:\n{json.dumps(schema, indent=4)}")
        try:
            jsonschema.validate(instance=to_validate, schema=schema)
        except jsonschema.exceptions.ValidationError as err:
            raise err
        self.logger.info("Schema is valid.")



import re
from collections import defaultdict
from typing import Dict

import yaml
from prance import ResolvingParser
from prance.convert import convert_spec

from roboswag.generate.models.definition import Definition, Property
from roboswag.generate.models.endpoint import Endpoint
from roboswag.generate.models.parameter import Parameter
from roboswag.generate.models.response import Response
from roboswag.generate.models.tag import Tag
from roboswag.generate.models.utils import get_python_type, pythonify_name


class APIModel:
    def __init__(self) -> None:
        self.name: str = ""
        self.tags: Dict[str, Tag] = {}
        self.definitions: Dict[str, Definition] = {}

    def parse_swagger(self, swagger):
        self.parse_info(swagger)
        self.parse_paths(swagger)
        self.parse_tags(swagger)
        self.parse_definitions(swagger)

    def parse_info(self, swagger):
        name = swagger["info"]["title"].replace(" ", "")
        self.name = name

    def get_unique_endpoint_name(self, path, method, method_body):
        if "operationId" in method_body:
            return pythonify_name(method_body["operationId"])
        # split on camelCase and / path
        parts = re.split("((?<=[a-z])(?=[A-Z])|/)", path)
        name = method
        for part in parts:
            part = re.sub("[/{}]", "", part)
            part = part.replace("-", "_")
            if not part:
                continue
            name += f"_{part.lower()}"
        return name

    @staticmethod
    def class_name_from_tag(method_body, **kwargs):
        return method_body["tags"][0].strip(" -_").title()

    @staticmethod
    def class_name_from_path(path, **kwargs):
        for part in path.split("/"):
            part = re.sub("[/{}]", "", part)
            if not part:
                continue
            return part.title()

    def set_source_of_class_name(self, swagger):
        """
        If all methods have their tags - use tag name. Otherwise use first part of the path.
        """
        # TODO configurable class name
        for path_body in swagger["paths"].values():
            for method in path_body.values():
                if "tag" not in method:
                    return self.class_name_from_path
        return self.class_name_from_tag

    def parse_paths(self, swagger):
        get_class_name = self.set_source_of_class_name(swagger)
        for path, path_body in swagger["paths"].items():
            method: str
            method_body: Dict
            for method, method_body in path_body.items():
                class_name = get_class_name(path=path, method_body=method_body)
                unique_name = self.get_unique_endpoint_name(path, method, method_body)
                summary = method_body.get("summary", "")
                description = method_body.get("description", "")
                params = defaultdict(list)
                for param in method_body.get("parameters", []):
                    params[param["in"]].append(
                        Parameter(
                            param["name"],
                            default="None" if param["in"] != "path" else None,  # TODO Retrieve default value
                            param_type=get_python_type(param["type"], param.get("format"))
                            if param.get("type")
                            else None,
                            description=param.get("description"),
                            required=param.get("required"),
                            schema=param.get("schema"),
                        )
                    )
                responses = dict()
                for status_code, resp in method_body.get("responses", []).items():
                    responses[status_code] = Response(
                        description=resp.get("description"),
                        headers=resp.get("headers"),
                        schema=resp.get("schema"),
                    )
                body = params["body"][0] if params["body"] else None
                endpoint = Endpoint(
                    unique_name,
                    method,
                    path,
                    summary=summary,
                    description=description,
                    path_params=params["path"],
                    headers=params["header"],
                    query=params["query"],
                    body=body,
                    responses=responses,
                )
                self.add_endpoint_to_tag(class_name, endpoint)

    def parse_tags(self, swagger):
        if not swagger.get("tags"):
            return
        for tag in swagger["tags"]:
            tag_name = Tag.normalize_tag_name(tag["name"])
            if tag_name in self.tags:
                self.tags[tag_name].description = tag["description"]

    def add_endpoint_to_tag(self, tag: str, endpoint: Endpoint) -> None:
        if tag not in self.tags:
            self.tags[tag] = Tag(tag)
        self.tags[tag].endpoints.append(endpoint)

    def parse_definitions(self, swagger):
        for def_name, def_body in swagger["definitions"].items():
            def_type = def_body.get("type", "")
            def_req = def_body.get("required")
            properties = []
            for prop_name, prop_body in def_body.get("properties", {}).items():
                prop_type = prop_body.get("type", "")
                prop_format = prop_body.get("format", "")
                properties.append(Property(prop_name, prop_type=get_python_type(prop_type, prop_format)))

            definition = Definition(def_name, def_type=def_type, properties=properties, required=def_req)
            self.definitions[def_name] = definition


class APIModelCreator:
    @staticmethod
    def from_yaml(source):
        with open(source) as f:
            data = yaml.load(f, Loader=yaml.Loader)
        api_model = APIModel()
        api_model.parse_swagger(data)
        return api_model, data

    @staticmethod
    def from_prance(source, convert_to_3=False):
        # TODO: support for swagger 3.0 (https://swagger.io/specification/)
        parser = ResolvingParser(source)
        swagger = parser.specification
        # convert to OpenAPI 3.x if swagger is in version 2.x
        if swagger.get("swagger") and convert_to_3:
            swagger = convert_spec(swagger).specification
        api_model = APIModel()
        api_model.parse_swagger(swagger)
        return api_model, swagger



from typing import List

from roboswag.generate.models.endpoint import Endpoint
from roboswag.generate.models.utils import pythonify_name


class Tag:
    def __init__(self, name: str, description: str = "") -> None:
        # tag is grouped paths/endpoints
        self.name: str = pythonify_name(name, join_mark="", join_fn="title")
        self.description: str = description
        self.endpoints: List[Endpoint] = []

    @staticmethod
    def normalize_tag_name(tag_name: str) -> str:
        return tag_name.strip(" -_").title() + "API"



import json
from pathlib import Path
from typing import List, Optional

import black
from jinja2 import Template

from roboswag.generate.models.api import APIModelCreator
from roboswag.generate.models.definition import Definition


def generate_libraries(source, output: Optional[Path] = None):
    api_model, swagger = APIModelCreator.from_prance(source)
    output_dir = api_model.name
    if output is not None:
        output_dir = output / output_dir
    Path(output_dir).mkdir(exist_ok=True)

    init_files = generate_init(swagger, output_dir)
    blackify_files(init_files)

    endpoints_dir = Path(output_dir) / Path("endpoints")
    endpoint_files = generate_endpoints(api_model.tags, endpoints_dir)
    blackify_files(endpoint_files)

    models_dir = Path(output_dir) / Path("models")
    model_files = generate_models(api_model.definitions, models_dir)
    blackify_files(model_files)

    schemas_dir = Path(output_dir) / Path("schemas")
    generate_schemas(swagger, schemas_dir)


def blackify_files(path_list: List[Path]):
    for path in path_list:
        blackify_file(path)


def blackify_file(source):
    black.format_file_in_place(source, fast=True, mode=black.FileMode(), write_back=black.WriteBack.YES)


def generate_init(swagger, output_dir):
    parent_dir = Path(__file__).parent
    Path(output_dir).mkdir(exist_ok=True)
    swagger_version = swagger.get("openapi") or swagger.get("swagger")
    with open(Path(parent_dir, "templates/api_init.jinja")) as f:
        template = Template(f.read()).render(swagger_version=swagger_version, infos=swagger["info"])
    init_file = Path(output_dir, f"__init__.py")
    with open(init_file, "w") as f:
        f.write(template)
    print(f"Generated '{init_file}' file")
    return [init_file]


def generate_endpoints(tags, output_dir):
    parent_dir = Path(__file__).parent
    Path(output_dir).mkdir(exist_ok=True)
    print("Generating endpoints...")
    endpoint_files = []
    for tag in tags.values():
        with open(Path(parent_dir, "templates/paths.jinja")) as f:
            template = Template(f.read()).render(
                class_name=tag.name,
                endpoints=tag.endpoints,
                description=tag.description,
            )
        endpoint_file = Path(output_dir, f"{tag.name}.py")
        with open(endpoint_file, "w") as f:
            f.write(template)
        print(f"Generated '{endpoint_file}' file")
        endpoint_files.append(endpoint_file)
    return endpoint_files


def generate_models(definitions, output_dir):
    parent_dir = Path(__file__).parent
    Path(output_dir).mkdir(exist_ok=True)
    definition: Definition
    print("Generating models...")
    model_files = []
    for definition in definitions.values():
        Path(output_dir).mkdir(exist_ok=True)
        with open(Path(parent_dir, "templates/models.jinja")) as f:
            template = Template(f.read()).render(class_name=definition.name, properties=definition.properties)
        model_file = Path(output_dir, f"{definition.name}.py")
        with open(model_file, "w") as f:
            f.write(template)
        print(f"Generated '{model_file}' file")
        model_files.append(model_file)
    return model_files


def generate_schemas(swagger, output_dir):
    Path(output_dir).mkdir(exist_ok=True)
    print("Generating schemas...")
    schema_files = []
    for schema_name, schema in swagger["definitions"].items():
        Path(output_dir).mkdir(exist_ok=True)
        schema_file = Path(output_dir, f"{schema_name}.json")
        with open(schema_file, "w") as f:
            f.write(json.dumps(schema, indent=4))
            f.write("\n")
        print(f"Generated '{schema_file}' file")
        schema_files.append(schema_file)
    return schema_files



import rich_click as click

from roboswag.generate import generate_libraries
from roboswag.version import __version__

CONTEXT_SETTINGS = {"help_option_names": ["-h", "--help"]}


@click.group(invoke_without_command=True, context_settings=CONTEXT_SETTINGS)
@click.version_option(version=__version__, prog_name="roboswag")
def cli():
    """
    Roboswag is a tool that generates Python libraries out of your Swagger (OpenAPI specification file).
    """
    pass


@cli.command()
@click.option(
    "-s",
    "--spec",
    type=click.Path(
        exists=True,
        file_okay=True,
        dir_okay=False,
        readable=True,
        allow_dash=False,
        path_type=str,
    ),
    required=True,
    metavar="SWAGGER",
    help="OpenAPI specification file",
)
def generate(spec: str):
    """Generate Python libraries."""
    generate_libraries(spec)



import datetime
import re
from typing import Dict, List

RESERVED_WORDS = {"global", "cls", "self"}

types_mapping = {
    "string": {
        "": str,
        "byte": str,
        "password": str,
        "date": datetime.date,
        "date-time": datetime.datetime,
        "binary": bytes,
    },
    "integer": {
        "": int,
        "int32": int,
        "int64": int,
    },
    "number": {
        "float": float,
        "double": float,
    },
    "boolean": bool,
    "file": "file",
    "array": list(),
    "object": object,
}


def get_python_type(param_type, param_format=None):
    if not param_type:
        return str
    if not types_mapping[param_type]:
        return str
    if not isinstance(types_mapping[param_type], Dict):
        return types_mapping[param_type]
    if param_format:
        return types_mapping[param_type][param_format]
    return str


def pythonify_name(name: str, join_mark: str = "_", join_fn: str = "lower") -> str:
    names = re.split("([A-Z][a-z]+)", name)
    if join_fn == "lower":
        name = join_mark.join(name.lower() for name in names if name.strip())
    elif join_fn == "title":
        name = join_mark.join(name.title() for name in names if name.strip())
    name = name.replace("-", "")
    return name


def replace_reserved_name(name: str) -> str:
    """If the name is in reserved words, append it with _"""
    if name in RESERVED_WORDS:
        return f"_{name}"
    return name



from roboswag.validate.core import ValidateBase


class ValidateError(ValidateBase):
    def error(self, response, exp_error):
        self.logger.info("Validating error message...")
        assertion_err_msg = (
            f"Received response description:\n    '{response.text}'\n" f"does not equal expected:\n    '{exp_error}'"
        )
        assert response.text == exp_error, assertion_err_msg
        self.logger.info("Error message is valid.")



from pprint import pformat

from robot.api import logger


class Logger:
    @staticmethod
    def info(message):
        logger.info(message)

    @staticmethod
    def debug(message):
        logger.debug(message)

    @staticmethod
    def log_request(response):
        # TODO we can add flag for removing auth details (such as passwords, tokens) if needed
        if response.history:  # TODO print redirects
            original_request = response.history[0].request
            redirected = "(redirected)"
        else:
            original_request = response.request
            redirected = ""
        logger.info(
            f"{original_request.method.upper()} {original_request.url} {redirected}\n"
            f"headers: {pformat(original_request.headers)}"
        )  # TODO make pretty print work

    @staticmethod
    def log_response(response):
        logger.info(f"{response.request.method.upper()} response: {response.status_code} {response.text}")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['roboswag',
 'roboswag.generate',
 'roboswag.generate.models',
 'roboswag.validate']

package_data = \
{'': ['*'], 'roboswag.generate': ['templates/*']}

install_requires = \
['Jinja2>=3.1.2,<4.0.0',
 'PyYAML>=6.0,<7.0',
 'black>=22.3.0,<23.0.0',
 'click>=8.1.0',
 'jsonschema>=4.5.1,<5.0.0',
 'openapi-spec-validator>=0.4.0,<0.5.0',
 'prance>=0.21.8,<0.22.0',
 'requests>=2.27.1,<3.0.0',
 'rich_click>=1.4,<2.0',
 'robotframework>=4.1',
 'urllib3>=1.26.9,<2.0.0']

entry_points = \
{'console_scripts': ['roboswag = roboswag:cli']}

setup_kwargs = {
    'name': 'roboswag',
    'version': '0.3.1',
    'description': 'Test framework for auto-generating libraries from OpenAPI specification file.',
    'long_description': "# Roboswag \n\n- [Introduction](#introduction)\n- [Installation](#installation)\n- [Usage](#usage)\n- [Limitations](#limitations)\n\n## Introduction\n\n:robot: Roboswag is a tool that automatically generates Python libraries out of your Swagger (OpenAPI specification \nfile). These libraries can be used to create tests with various payload content and its validation. It also supports \nresponse validation against schema and verification of different status codes.\n\n> **Note**\n>\n> The tool is in the ***Alpha*** state, which means it may be unstable and should be used at your own risk. Some \n> features may be broken and there are still many things to be developed. Happy testing!\n\nThe OpenAPI Specification (OAS) defines a standard, language-agnostic interface to RESTful APIs.\nClick [here for v3 documentation](https://swagger.io/specification/) and\n[here for v2 documentation](https://swagger.io/specification/v2).\n\n> Hosted on [GitHub](https://github.com/MarketSquare/roboswag). :medal_military:\n\n## Installation\n\n> **Note**\n>\n> The PyPI package will be released when Beta version is out.\n\nFirst, you need to install Roboswag, and there are currently 2 ways to do it:\n- you can clone the repository, go to the main `roboswag` directory and install the tool locally:\n\n```commandline\npip install .\n```\n\n- you can install the tool directly from GitHub's source code:\n\n```commandline\npip install -U git+https://github.com/MarketSquare/roboswag.git\n```\n\n## Usage\n\nRoboswag can be easily run from command line. To check if it's installed, run this to see the current version:\n\n```commandline\nroboswag -v\n```\n\nTo execute Roboswag with its full capabilities, run it with provided path to the Swagger (OpenAPI specification) file:\n\n```commandline\nroboswag generate -s <path_to_swagger>\n```\n\n> You can try out the tool using the example of swagger file located in `swaggers/petstore.json`.\n\nSuccessful execution should result in printing the information about generated files and a whole new directory (named \nby the value of `info.title` from your Swagger file) consisting of:\n- `endpoints` directory with files representing each `tag` as a class with methods representing its endpoints,\n- `models` directory with API models represented as Python classes,\n- `schemas` directory with every possible schema from your API as JSON file used for validating payload and responses. \n\nNow you can just create a test file, import desired endpoint and start automating the testing!\n\n## Limitations\n\nThe tool is already able to generate libraries but...\n- Not all fields from the swagger specification may be supported. This means that a specific file may break the tool \n  and flood the terminal with stack trace\n- Also, the support for Swagger V3 is not yet implemented\n- Authorization to access the API is not yet fully covered\n- There is not much to be configured here - it works always the same\n- There is no real documentation apart from this file\n- There are nearly no tests assuring this tool works correctly\n\nPlease be forgiving and submit an issue, if you struggle with something or just contact us on our\n[Slack channel](https://robotframework.slack.com/archives/C035KMZ2FGA). It's more than welcome also to support us by \ncode contribution! :keyboard:\n",
    'author': 'Mateusz Nojek',
    'author_email': 'matnojek@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://pypi.org/project/roboswag',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from typing import List

from roboswag.generate.models.utils import get_python_type, pythonify_name, replace_reserved_name


class Property:
    def __init__(self, name: str, prop_type=None):
        self.name: str = replace_reserved_name(pythonify_name(name))
        self.type = prop_type


class Definition:
    def __init__(self, name: str, def_type, properties: List[Property], required=None):
        self.name: str = name
        self.type = get_python_type(def_type)
        self.required = required
        self.properties: List[Property] = properties
